from __future__ import annotations

from typing import Optional
from datetime import datetime

from pydantic import BaseModel, Field, validator

from sankaku import types
from sankaku.utils import convert_ts_to_datetime
from .tags import PostTag
from .users import Author


__all__ = ["Comment", "Post", "AIPost"]


class GenerationDirectives(BaseModel):
    """Model that describes additional fields for AI-generated posts."""
    width: int
    height: int
    prompt: str
    batch_size: int
    batch_count: int
    sampling_steps: int
    negative_prompt: str


class BasePost(BaseModel):
    """Model that contains minimum amount of information that all posts have."""
    id: int
    created_at: datetime
    rating: types.Rating
    status: str
    author: Author
    file_url: Optional[str]
    preview_url: Optional[str]
    width: int
    height: int
    file_size: int
    extension: Optional[str] = Field(alias="file_type")
    generation_directives: Optional[GenerationDirectives]
    md5: str
    tags: list[PostTag]

    @property
    def file_type(self) -> Optional[types.FileType]:
        """Get type of the file."""
        match self.extension:
            case "png" | "jpeg" | "webp":
                return types.FileType.IMAGE
            case "webm" | "mp4":
                return types.FileType.VIDEO
            case "gif":
                return types.FileType.GIF
            case _:
                return None

    # Validators
    _normalize_datetime = (
        validator("created_at", pre=True, allow_reuse=True)
        (convert_ts_to_datetime)
    )

    @validator("extension", pre=True)
    def get_extension(cls, v) -> Optional[str]:  # noqa
        return v.split("/")[-1] if v else None


class Comment(BaseModel):
    """Model that describes comments related to posts if they are exist."""
    id: int
    created_at: datetime
    post_id: int
    author: Author
    body: str
    score: int
    parent_id: Optional[int]
    children: list[Comment]
    deleted: bool
    deleted_by: dict  # Seen only empty dictionaries so IDK real type
    updated_at: Optional[datetime]
    can_reply: bool
    reason: None  # Seen only None values so IDK real type


class Post(BasePost):
    """Model that describes posts."""
    sample_url: Optional[str]
    sample_width: int
    sample_height: int
    preview_width: Optional[int]
    preview_height: Optional[int]
    has_children: bool
    has_comments: bool
    has_notes: bool
    is_favorited: bool
    user_vote: Optional[int]
    parent_id: Optional[int]
    change: Optional[int]
    fav_count: int
    recommended_posts: int
    recommended_score: int
    vote_count: int
    total_score: int
    comment_count: Optional[int]
    source: Optional[str]
    in_visible_pool: bool
    is_premium: bool
    is_rating_locked: bool
    is_note_locked: bool
    is_status_locked: bool
    redirect_to_signup: bool
    sequence: Optional[int]
    video_duration: Optional[float]


class AIPost(BasePost):
    """Model that describes AI-generated posts.

    There is possibility that AI posts have the same fields as common posts,
    but premium account is needed to check it properly. So this model is
    actual for non-premium accounts.
    """
    updated_at: Optional[datetime]
    post_associated_id: Optional[int]

    # Validators
    _normalize_datetime = (
        validator("created_at", "updated_at", pre=True, allow_reuse=True)
        (convert_ts_to_datetime)
    )



from typing import Generic, TypeVar
from dataclasses import dataclass


__all__ = ["Page"]


_T = TypeVar("_T")


@dataclass()
class Page(Generic[_T]):
    """Model that describes page containing content with specific type."""
    number: int
    items: list[_T]



from dataclasses import dataclass
from typing import TypedDict, Optional


__all__ = ["ValueRange", "Timestamp"]


@dataclass(frozen=True)
class ValueRange:
    min: int
    max: int


class Timestamp(TypedDict):
    json_class: str
    s: Optional[int]
    n: int



from datetime import datetime
from typing import Optional

from pydantic import BaseModel, Field, validator

from sankaku import types


__all__ = ["Author", "User", "ExtendedUser"]


class BaseUser(BaseModel):
    """User profile with a minimum amount of information."""
    id: int
    name: str
    avatar: str
    avatar_rating: types.Rating


class Author(BaseUser):
    """Model that describes users who are the authors of posts or wiki pages."""


class User(BaseUser):
    """User profile model for any user that has an account on website."""

    # The following fields are always present
    level: int
    upload_limit: int
    created_at: datetime
    favs_are_private: bool
    avatar: str = Field(alias="avatar_url")
    post_upload_count: int
    pool_upload_count: int
    comment_count: int
    post_update_count: int
    note_update_count: int
    wiki_update_count: int
    forum_post_count: int
    pool_update_count: int
    series_update_count: int
    tag_update_count: int
    artist_update_count: int
    show_popup_version: int
    credits: int
    credits_subs: int
    is_ai_beta: bool

    # The following fields can be missing in API json response
    last_logged_in_at: Optional[datetime] = None
    favorite_count: Optional[int] = None
    post_favorite_count: Optional[int]
    pool_favorite_count: Optional[int]
    vote_count: Optional[int] = None
    post_vote_count: Optional[int] = None
    pool_vote_count: Optional[int] = None
    recommended_posts_for_user: Optional[int] = None
    subscriptions: list[str] = []


class ExtendedUser(User):
    """Profile of the currently logged-in user."""
    email: str
    hide_ads: bool
    subscription_level: int
    filter_content: bool
    receive_dmails: bool
    email_verification_status: str
    is_verified: bool
    verifications_count: int
    blacklist_is_hidden: bool
    blacklisted_tags: list[str]
    blacklisted: list[str]
    mfa_method: int

    @validator("blacklisted_tags", pre=True)
    def flatten_blacklisted_tags(cls, v) -> list[str]:  # noqa
        """Flatten nested lists into one."""
        return [tag[0] for tag in v]



from typing import Optional
from datetime import datetime

from pydantic import BaseModel, Field, validator

from sankaku import types
from sankaku.utils import convert_ts_to_datetime
from .users import Author


__all__ = ["PostTag", "PageTag", "Wiki", "WikiTag"]


class BaseTag(BaseModel):
    """Model that contains minimum amount of information that all tags have."""
    id: int
    name: str
    name_en: str
    name_ja: Optional[str]
    type: types.TagType
    post_count: int
    pool_count: int
    series_count: int
    rating: Optional[types.Rating]


class TagMixin(BaseModel):
    """Additional data that certain tags have."""
    count: int
    tag_name: str = Field(alias="tagName")
    total_post_count: int
    total_pool_count: int


class PostTag(BaseTag, TagMixin):
    """Model that describes tags related to posts."""
    locale: str
    version: Optional[int]


class NestedTag(BaseTag):
    """Model that describes tags with specific relation to certain tag on tag page."""
    post_count: int = Field(alias="postCount")
    cached_related: Optional[list[int]] = Field(alias="cachedRelated")
    cached_related_expires_on: datetime = Field(alias="cachedRelatedExpiresOn")
    type: types.TagType = Field(alias="tagType")
    name_en: str = Field(alias="nameEn")
    name_ja: Optional[str] = Field(alias="nameJa")
    popularity_all: Optional[float] = Field(alias="scTagPopularityAll")
    quality_all: Optional[float] = Field(alias="scTagQualityAll")
    popularity_ero: Optional[float] = Field(alias="scTagPopularityEro")
    popularity_safe: Optional[float] = Field(alias="scTagPopularitySafe")
    quality_ero: Optional[float] = Field(alias="scTagQualityEro")
    quality_safe: Optional[float] = Field(alias="scTagQualitySafe")
    parent_tags: Optional[list[int]] = Field(alias="parentTags")
    child_tags: Optional[list[int]] = Field(alias="childTags")
    pool_count: int = Field(alias="poolCount")
    premium_post_count: int = Field(alias="premPostCount")
    non_premium_post_count: int = Field(alias="nonPremPostCount")
    premium_pool_count: int = Field(alias="premPoolCount")
    non_premium_pool_count: int = Field(alias="nonPremPoolCount")
    series_count: int = Field(alias="seriesCount")
    premium_series_count: int = Field(alias="premSeriesCount")
    non_premium_series_count: int = Field(alias="nonPremSeriesCount")
    is_trained: bool = Field(alias="isTrained")
    child: int
    parent: int
    version: Optional[int]

    @validator("cached_related", "parent_tags", "child_tags", pre=True)
    def flatten(cls, v) -> Optional[list[int]]:  # noqa
        """Flatten nested lists into one."""
        if not v:
            return None
        tag_ids = v.split(",") if "," in v else v.split()
        try:
            return [int(tag_id) for tag_id in tag_ids]
        except ValueError:
            return None


class Translations(BaseModel):
    """Model that describes tag translations if they are present."""
    root_id: int = Field(alias="rootId")
    lang: str
    translation: str


class PageTag(PostTag):
    """Model that describes tags on tag page."""
    translations: list[Translations]
    related_tags: list[NestedTag]
    child_tags: list[NestedTag]
    parent_tags: list[NestedTag]


class Wiki(BaseModel):
    """Model that describes wiki information for specific tag."""
    id: int
    title: str
    body: str
    created_at: datetime
    updated_at: Optional[datetime]
    author: Author = Field(alias="user")
    is_locked: bool
    version: int

    # Validators
    _normalize_datetime = (
        validator("created_at", "updated_at", pre=True, allow_reuse=True)
        (convert_ts_to_datetime)
    )


class WikiTag(BaseTag, TagMixin):
    """Model that describes tag on wiki page."""
    related_tags: list[PostTag]
    child_tags: list[PostTag]
    parent_tags: list[PostTag]
    alias_tags: list[PostTag]
    implied_tags: list[PostTag]
    translations: list[Translations]
    wiki: Wiki



from aiohttp import ClientSession
from loguru import logger

from .abc import ABCHttpClient
from sankaku import errors, constants as const
from sankaku.models.http import ClientResponse


__all__ = ["HttpClient"]


class HttpClient(ABCHttpClient):
    """HTTP client for API requests that instances use a single session."""
    def __init__(self) -> None:
        self.headers: dict[str, str] = const.HEADERS.copy()
        self.session: ClientSession = ClientSession()

    def __del__(self) -> None:
        if not self.session.closed and self.session.connector is not None:
            self.session.connector.close()

    async def close(self) -> None:
        """There is no need to close client with single session."""

    async def request(self, method: str, url: str, **kwargs) -> ClientResponse:
        """Make request to specified url."""
        if kwargs.get("headers") is None:
            kwargs["headers"] = self.headers

        response = await self.session.request(method, url, **kwargs)
        logger.debug(f"Sent {method} request to {response.url}")

        if response.content_type != "application/json":
            raise errors.SankakuServerError(
                response.status, "Invalid response content type",
                content_type=response.content_type
            )

        client_response = ClientResponse(
            response.status,
            await response.json(encoding="utf-8"),
        )
        response.close()
        logger.debug(
            f"Request {method} returned response with status "
            f"[{client_response.status}]: {client_response.json}",
        )

        return client_response

    async def get(self, url: str, **kwargs) -> ClientResponse:
        """Send GET request to specified url."""
        return await self.request("GET", url, **kwargs)

    async def post(self, url: str, **kwargs) -> ClientResponse:
        """Send POST request to specified url."""
        return await self.request("POST", url, **kwargs)



from datetime import datetime

import pytest  # noqa

from sankaku import utils, errors


@pytest.mark.parametrize(
    ["rps", "rpm", "expected"],
    [(200, 200, errors.RateLimitError), (None, None, TypeError)]
)
async def test_ratelimit_with_incompatible_args(rps, rpm, expected):
    with pytest.raises(expected):
        @utils.ratelimit(rps=rps, rpm=rpm)
        async def idle_request():
            assert True

        await idle_request()


@pytest.mark.parametrize(["rps", "rpm"], [(3, None), (180, None)])
async def test_ratelimit(rps, rpm):
    @utils.ratelimit(rps=rps, rpm=rpm)
    async def idle_request():
        assert True

    await idle_request()


@pytest.mark.parametrize(
    ["ts", "expected"],
    [
        (
            {
                "json_class": "Time",
                "s": 1680471860,
                "n": 0
            },
            datetime.utcfromtimestamp(1680471860).astimezone()
        ),
        (
            {
                "json_class": "Time",
                "s": None,
                "n": 0
            },
            None
        ),
    ]
)
def test_convert_ts_to_datetime(ts, expected):
    assert utils.convert_ts_to_datetime(ts) == expected



"""Miscellaneous support functions that are used at different places."""

import asyncio
from datetime import datetime
from functools import wraps
from typing import TypeVar, ParamSpec, Optional, Any
from collections.abc import Callable, Awaitable


from sankaku.errors import RateLimitError
from sankaku.typedefs import Timestamp


__all__ = ["ratelimit", "convert_ts_to_datetime", "from_locals"]


_T = TypeVar("_T")
_P = ParamSpec("_P")


def ratelimit(
    *,
    rps: Optional[int] = None,
    rpm: Optional[int] = None
) -> Callable[[Callable[_P, Awaitable[_T]]], Callable[_P, Awaitable[_T]]]:
    """
    Limit the number of requests.

    :param rps: Request per second
    :param rpm: Requests per minute
    """
    if all(locals().values()):
        raise RateLimitError
    elif not any(locals().values()):
        raise TypeError("At least one argument must be specified.")

    sleep_time: float = (1 / rps) if rps else (60 / rpm)  # type: ignore[operator]

    def wrapper(func: Callable[_P, Awaitable[_T]]) -> Callable[_P, Awaitable[_T]]:
        @wraps(func)
        async def inner(*args: _P.args, **kwargs: _P.kwargs) -> _T:
            await asyncio.sleep(sleep_time)
            return await func(*args, **kwargs)
        return inner

    return wrapper


def convert_ts_to_datetime(ts: Timestamp) -> Optional[datetime]:
    """Convert timestamp in datetime dict into datetime class."""
    if ts.get("s") is None:
        return None
    return datetime.utcfromtimestamp(ts["s"]).astimezone()  # type: ignore[arg-type]


def from_locals(
    loc: dict[str, Any], exclude: tuple[str, ...] = ("self",)
) -> dict[str, Any]:
    """Get arguments of calling function from its locals to pass them to paginator.

    :param loc: locals of calling function
    :param exclude: arguments to be excluded
    """
    return {k: v for k, v in loc.copy().items() if k not in exclude}



from typing import Optional


__all__ = [
    "SankakuError",
    "RateLimitError",
    "LoginRequirementError",
    "VideoDurationError",
    "SankakuServerError",
    "PaginatorLastPage",
    "PageNotFoundError",
    "AuthorizationError"
]


class SankakuError(Exception):
    """Base error class for raising exceptions without any special params."""
    msg: str = ""

    def __init__(self, msg: Optional[str] = None) -> None:
        self.msg = msg or self.msg

    def __repr__(self) -> str:
        return repr(self.msg)

    def __str__(self) -> str:
        return str(self.msg)


class RateLimitError(SankakuError):
    msg = "Can't set both rps and rpm at once."


class LoginRequirementError(SankakuError):
    msg = "You must be logged-in."


class VideoDurationError(SankakuError):
    msg = "Argument is available only with video files."


class SankakuServerError(SankakuError):
    """Error class for parametrized exceptions."""
    def __init__(
            self,
            status: Optional[int],
            msg: Optional[str] = None,
            **kwargs
    ) -> None:
        self.status = status
        self.kwargs = kwargs

        str_kwargs = ", ".join(f"{k}={v}" for k, v in self.kwargs.items())
        delimiter = ": " if self.kwargs else ""
        self.msg = f"[{self.status}] {msg or self.msg}{delimiter}{str_kwargs}."

    def __repr__(self) -> str:
        return repr(self.msg)

    def __str__(self) -> str:
        return str(self.msg)


class PaginatorLastPage(SankakuServerError):
    msg = "Last available page reached"


class PageNotFoundError(SankakuServerError):
    msg = f"Failed to fetch page with requested params"


class AuthorizationError(SankakuServerError):
    msg = f"Authorization failed"



from datetime import datetime

import pytest  # noqa

from sankaku import errors, models as mdl, types
from sankaku.clients import SankakuClient
from sankaku.utils import from_locals


class TestBaseClient:
    @pytest.mark.parametrize(
        ["data", "expected"],
        [
            ({"access_token": "invalid"}, errors.AuthorizationError),
            ({"login": "invalid", "password": "invalid"}, errors.AuthorizationError),
            ({}, errors.SankakuError)
        ]
    )
    async def test_login_with_invalid_data(
            self, nlclient: SankakuClient,
            data, expected
    ):
        with pytest.raises(expected):
            await nlclient.login(**data)

    async def test_login_with_valid_data(self, lclient: SankakuClient):
        assert isinstance(lclient.profile, mdl.ExtendedUser)


class TestPostClient:
    async def test_browse_default(self, nlclient: SankakuClient):
        """Default behaviour when unauthorized user don't set any arguments."""
        assert isinstance(await anext(nlclient.browse_posts()), mdl.Post)

    @pytest.mark.parametrize(
        ["file_type", "video_duration", "expected"],
        [(types.FileType.IMAGE, [1, 60], errors.VideoDurationError)]
    )
    async def test_browse_with_incompatible_args(
        self, nlclient: SankakuClient,
        file_type, video_duration, expected
    ):
        """Case when arguments are incompatible."""
        with pytest.raises(expected):
            await anext(
                nlclient.browse_posts(
                    file_type=file_type, video_duration=video_duration
                )
            )

    @pytest.mark.parametrize(["page_number", "limit"], [(-1, 40), (1, -40)])
    async def test_browse_with_incorrect_page_number_or_limit(
        self, nlclient: SankakuClient,
        page_number, limit
    ):
        with pytest.raises(errors.SankakuServerError):
            async for _ in nlclient.browse_posts(
                page_number=page_number, limit=limit
            ):
                break

    @pytest.mark.parametrize(
        [
            "page_number", "limit", "hide_posts_in_books", "order",
            "date", "rating", "threshold", "file_size", "file_type",
            "video_duration", "recommended_for", "favorited_by", "tags",
            "added_by", "voted"

        ],
        [
            (
                1, 50, "always", types.PostOrder.POPULARITY,
                None, None, None, None, None,
                None, None, None, None,
                None, None
            ),
            (
                1, 40, "in-larger-tags", types.PostOrder.QUALITY,
                None, types.Rating.EXPLICIT, None, None, types.FileType.IMAGE,
                None, "Nigredo", None, None,
                None, None
            ),
            (
                1, 10, None, types.PostOrder.DATE,
                [datetime(2018, 6, 16)], None, 4, None, types.FileType.VIDEO,
                [1, 900], None, None, ["animated"],
                ["anonymous"], "Nigredo"
            ),
            (
                1, 10, None, types.PostOrder.DATE,
                None, None, 4, types.FileSize.LARGE, None,
                None, None, "Nigredo", ["female", "solo"],
                None, None
            )
        ]
    )
    async def test_browse_with_random_args(
        self, lclient: SankakuClient,
        page_number, limit, hide_posts_in_books, order,
        date, rating, threshold, file_size, file_type,
        video_duration, recommended_for, favorited_by, tags,
        added_by, voted
    ):
        post = await anext(
            lclient.browse_posts(**from_locals(locals(), ("self", "lclient")))
        )
        assert isinstance(post, mdl.Post)

    async def test_get_favorited_posts_unauthorized(self, nlclient: SankakuClient):
        with pytest.raises(errors.LoginRequirementError):
            await anext(nlclient.get_favorited_posts())

    async def test_get_favorited_posts_authorized(self, lclient: SankakuClient):
        assert isinstance(await anext(lclient.get_favorited_posts()), mdl.Post)

    async def test_get_top_posts(self, nlclient: SankakuClient):
        assert isinstance(await anext(nlclient.get_top_posts()), mdl.Post)

    async def test_get_popular_posts(self, nlclient: SankakuClient):
        assert isinstance(await anext(nlclient.get_popular_posts()), mdl.Post)

    async def test_get_recommended_posts_unauthorized(self, nlclient: SankakuClient):
        with pytest.raises(errors.LoginRequirementError):
            await anext(nlclient.get_recommended_posts())

    async def test_get_recommended_posts_authorized(self, lclient: SankakuClient):
        assert isinstance(await anext(lclient.get_recommended_posts()), mdl.Post)

    @pytest.mark.parametrize(["post_id"], [(32948875,)])
    async def test_get_similar_posts(self, lclient: SankakuClient, post_id):
        assert isinstance(await anext(lclient.get_similar_posts(post_id)), mdl.Post)

    @pytest.mark.parametrize(["post_id"], [(33108291,)])
    async def test_get_post_comments(self, lclient: SankakuClient, post_id):
        assert isinstance(await anext(lclient.get_post_comments(post_id)), mdl.Comment)

    @pytest.mark.parametrize(["post_id"], [(32948875,), (33108291,)])
    async def test_get_post(self, nlclient: SankakuClient, post_id):
        post = await nlclient.get_post(post_id)
        assert isinstance(post, mdl.Post)

    async def test_get_non_existent_post(self, lclient: SankakuClient):
        with pytest.raises(errors.PageNotFoundError):
            await lclient.get_post(-10_000)

    async def test_create_post(self, lclient: SankakuClient):
        with pytest.raises(NotImplementedError):
            await lclient.create_post()


class TestAIClient:
    async def test_browse_default(self, nlclient: SankakuClient):
        """Default behaviour when unauthorized user don't set any arguments."""
        assert isinstance(await anext(nlclient.browse_ai_posts()), mdl.AIPost)

    @pytest.mark.parametrize(["page_number", "limit"], [(-1, 50), (1, -50)])
    async def test_browse_with_incorrect_page_number_or_limit(
        self, nlclient: SankakuClient,
        page_number, limit
    ):
        with pytest.raises(errors.SankakuServerError):
            async for _ in nlclient.browse_ai_posts(
                page_number=page_number, limit=limit
            ):
                break

    @pytest.mark.parametrize(["post_id"], [(123,), (1721,)])
    async def test_get_ai_post(self, nlclient: SankakuClient, post_id):
        post = await nlclient.get_ai_post(post_id,)
        assert isinstance(post, mdl.AIPost)

    async def test_get_non_existent_ai_post(self, lclient: SankakuClient):
        with pytest.raises(errors.PageNotFoundError):
            await lclient.get_ai_post(-10_000)

    async def test_create_ai_post(self, lclient: SankakuClient):
        with pytest.raises(NotImplementedError):
            await lclient.create_ai_post()


class TestTagClient:
    async def test_browse_default(self, nlclient: SankakuClient):
        """Default behaviour when unauthorized user don't set any arguments."""
        assert isinstance(await anext(nlclient.browse_tags()), mdl.PageTag)

    @pytest.mark.parametrize(["page_number", "limit"], [(-1, 22), (1, -86)])
    async def test_browse_with_incorrect_page_number_or_limit(
        self, nlclient: SankakuClient,
        page_number, limit
    ):
        with pytest.raises(errors.SankakuServerError):
            async for _ in nlclient.browse_tags(
                page_number=page_number, limit=limit
            ):
                break

    @pytest.mark.parametrize(
        [
            "page_number", "limit", "tag_type",
            "order", "rating", "max_post_count",
            "sort_parameter", "sort_direction"
        ],
        [
            (
                7, 20, types.TagType.CHARACTER,
                None, types.Rating.SAFE, None,
                None, types.SortDirection.DESC
            ),
            (
                8, 60, types.TagType.COPYRIGHT,
                types.TagOrder.POPULARITY, types.Rating.QUESTIONABLE, None,
                None, types.SortDirection.ASC
            ),
            (
                18, 10, None,
                None, types.Rating.QUESTIONABLE, 2_537_220,
                types.SortParameter.NAME, types.SortDirection.ASC
            )
        ]
    )
    async def test_browse_with_random_args(
        self, lclient: SankakuClient,
        page_number, limit, tag_type,
        order, rating, max_post_count,
        sort_parameter, sort_direction
    ):
        tag = await anext(
            lclient.browse_tags(**from_locals(locals(), ("self", "lclient")))
        )
        assert isinstance(tag, mdl.PageTag)

    @pytest.mark.parametrize(
        ["name_or_id"], [("animated",), (100,)]
    )
    async def test_get_tag(self, lclient: SankakuClient, name_or_id):
        wiki_tag = await lclient.get_tag(name_or_id)
        assert isinstance(wiki_tag, mdl.WikiTag)

    async def test_get_non_existent_tag(self, lclient: SankakuClient):
        with pytest.raises(errors.PageNotFoundError):
            await lclient.get_tag(-10_000)


class TestBookClient:
    async def test_browse_default(self, nlclient: SankakuClient):
        assert isinstance(await anext(nlclient.browse_books()), mdl.PageBook)

    @pytest.mark.parametrize(["page_number", "limit"], [(-3, 40), (1, -22)])
    async def test_browse_with_incorrect_page_number_or_limit(
            self, nlclient: SankakuClient,
            page_number, limit
    ):
        with pytest.raises(errors.SankakuServerError):
            async for _ in nlclient.browse_books(
                page_number=page_number, limit=limit
            ):
                break

    @pytest.mark.parametrize(
        [
            "order", "rating", "recommended_for",
            "favorited_by", "tags", "added_by",
            "voted", "page_number", "limit"
        ],
        [
            (
                types.BookOrder.RANDOM, types.Rating.EXPLICIT, "reichan",
                None, None, None,
                None, 1, 50
            ),
            (
                types.BookOrder.POPULARITY, None, None,
                "Nigredo", None, None,
                "Nigredo", None, None
            ),
            (
                None, None, None,
                None, ["genshin_impact"], ["yanququ"],
                None, 1, 10
            )
        ]
    )
    async def test_browse_with_random_args(
            self, lclient: SankakuClient,
            order, rating, recommended_for,
            favorited_by, tags, added_by,
            voted, page_number, limit
    ):
        book = await anext(
            lclient.browse_books(**from_locals(locals(), ("self", "lclient")))
        )
        assert isinstance(book, mdl.PageBook)

    async def test_favorited_books_unauthorized(self, nlclient: SankakuClient):
        with pytest.raises(errors.LoginRequirementError):
            await anext(nlclient.get_favorited_books())

    async def test_favorited_books_authorized(self, lclient: SankakuClient):
        assert isinstance(await anext(lclient.get_favorited_books()), mdl.PageBook)

    async def test_recommended_books_unauthorized(self, nlclient: SankakuClient):
        with pytest.raises(errors.LoginRequirementError):
            await anext(nlclient.get_recommended_books())

    async def test_recommended_books_authorized(self, lclient: SankakuClient):
        assert isinstance(await anext(lclient.get_recommended_books()), mdl.PageBook)

    async def test_recently_read_books_unauthorized(self, nlclient: SankakuClient):
        with pytest.raises(errors.LoginRequirementError):
            await anext(nlclient.get_recently_read_books())

    async def test_recently_read_books_authorized(self, lclient: SankakuClient):
        assert isinstance(await anext(lclient.get_recently_read_books()), mdl.PageBook)

    @pytest.mark.parametrize(["post_id"], [(27038477,)])
    async def test_get_related_books(self, lclient: SankakuClient, post_id):
        assert isinstance(await anext(lclient.get_related_books(post_id)), mdl.PageBook)

    @pytest.mark.parametrize(["book_id"], [(1000,)])
    async def test_get_book(self, nlclient: SankakuClient, book_id):
        book = await nlclient.get_book(book_id)
        assert isinstance(book, mdl.Book)

    async def test_get_non_existent_book(self, lclient: SankakuClient):
        with pytest.raises(errors.PageNotFoundError):
            await lclient.get_book(-10_000)


class TestUserClient:
    async def test_browse_users(self, nlclient: SankakuClient):
        """Default behaviour when unauthorized user don't set any arguments."""

        assert isinstance(await anext(nlclient.browse_users()), mdl.User)
        assert isinstance(
            await anext(nlclient.browse_users(level=types.UserLevel.MEMBER)),
            mdl.User
        )

    @pytest.mark.parametrize(["page_number", "limit"], [(-3, 22), (1, -69)])
    async def test_browse_with_incorrect_page_number_or_limit(
        self, nlclient: SankakuClient,
        page_number, limit
    ):
        with pytest.raises(errors.SankakuServerError):
            async for _ in nlclient.browse_users(
                page_number=page_number, limit=limit
            ):
                break

    @pytest.mark.parametrize(["name_or_id"], [("anonymous",), (1423490,)])
    async def test_get_user(self, nlclient: SankakuClient, name_or_id):
        assert isinstance(await nlclient.get_user(name_or_id), mdl.User)

    @pytest.mark.parametrize(["name_or_id"], [("!@#sdcvjkj|",), (-1000,)])
    async def test_get_user_with_wrong_name_or_id(
            self, nlclient: SankakuClient, name_or_id
    ):
        with pytest.raises(errors.PageNotFoundError):
            await nlclient.get_user(name_or_id)



import json
from typing import Optional, Literal, Annotated
from collections.abc import AsyncIterator
from datetime import datetime

from loguru import logger

from .abc import ABCClient
from .http_client import HttpClient
from sankaku.typedefs import ValueRange
from sankaku.utils import from_locals
from sankaku import models as mdl, constants as const, types, errors
from sankaku.paginators import *


__all__ = [
    "PostClient",
    "AIClient",
    "TagClient",
    "BookClient",
    "UserClient",
]


class BaseClient(ABCClient):
    """Base client used for login."""
    def __init__(self) -> None:
        self._profile: Optional[mdl.ExtendedUser] = None
        self._http_client: HttpClient = HttpClient()
        self._access_token: Optional[str] = None  # TODO: ability to update access token
        self._token_type: Optional[str] = None

    async def _login_via_credentials(self, login: str, password: str) -> None:
        response = await self._http_client.post(
            const.LOGIN_URL,
            data=json.dumps({"login": login, "password": password})
        )

        if not response.ok:
            raise errors.AuthorizationError(response.status, **response.json)

        self._access_token = response.json["access_token"]
        self._token_type = response.json["token_type"]
        self._profile = mdl.ExtendedUser(**response.json["current_user"])

    async def _login_via_access_token(self, access_token: str) -> None:
        try:
            self._profile = await self._get_profile(access_token)
            # Update access token and token type after successful profile fetch
            self._access_token = access_token
            self._token_type = const.DEFAULT_TOKEN_TYPE
        except errors.SankakuServerError as e:
            raise errors.AuthorizationError(e.status, **e.kwargs)

    async def _get_profile(self, access_token: str) -> mdl.ExtendedUser:
        """Get user profile information from Sankaku server by access token."""
        if self._profile is not None:
            return self._profile

        headers = {"authorization": f"{const.DEFAULT_TOKEN_TYPE} {access_token}"}
        headers.update(self._http_client.headers)
        response = await self._http_client.get(f"{const.PROFILE_URL}", headers=headers)

        if not response.ok:
            raise errors.SankakuServerError(
                response.status, "Failed to get user profile", **response.json
            )

        return mdl.ExtendedUser(**response.json["user"])

    async def login(
        self,
        *,
        access_token: Optional[str] = None,
        login: Optional[str] = None,
        password: Optional[str] = None
    ) -> None:
        """Login into sankakucomplex.com via access token or credentials.
        In case when all arguments are specified, preference will be given
        to authorization by credentials.

        :param access_token: User access token
        :param login: User email or nickname
        :param password: User password
        """
        match (access_token, login, password):
            case [str(), str(), str()] | [_, str(), str()]:
                await self._login_via_credentials(login, password)  # type: ignore[arg-type]
            case [str(), _, _]:
                await self._login_via_access_token(access_token)  # type: ignore[arg-type]
            case _:
                raise errors.SankakuError(
                    "The given data is not enough "
                    "or invalid (perhaps of the wrong type)."
                )

        self._http_client.headers.update(
            authorization=f"{self._token_type} {self._access_token}"
        )
        logger.info(f"Successfully logged in as {self._profile.name}.")  # type: ignore[union-attr]

    @property
    def profile(self) -> Optional[mdl.ExtendedUser]:
        return self._profile


class PostClient(BaseClient):
    """Client for post browsing."""
    async def browse_posts(
        self,
        order: Optional[types.PostOrder] = None,
        date: Optional[list[datetime]] = None,
        rating: Optional[types.Rating] = None,
        threshold: Optional[Annotated[int, ValueRange(1, 100)]] = None,
        hide_posts_in_books: Optional[Literal["in-larger-tags", "always"]] = None,
        file_size: Optional[types.FileSize] = None,
        file_type: Optional[types.FileType] = None,
        video_duration: Optional[list[int]] = None,
        recommended_for: Optional[str] = None,
        favorited_by: Optional[str] = None,
        tags: Optional[list[str]] = None,
        added_by: Optional[list[str]] = None,
        voted: Optional[str] = None,
        *,
        page_number: Optional[int] = None,
        limit: Optional[Annotated[int, ValueRange(1, 100)]] = None

    ) -> AsyncIterator[mdl.Post]:
        """Get posts from post pages.

        :param order: Post order rule
        :param date: Date or range of dates
        :param rating: Post rating
        :param threshold: Vote (quality) filter of posts
        :param hide_posts_in_books: Whether show post from books or not
        :param file_size: Size (aspect ratio) of mediafile
        :param file_type: Type of mediafile in post
        :param video_duration: Video duration in seconds or in range of seconds
        :param recommended_for: Posts recommended for specified user
        :param favorited_by: Posts favorited by specified user
        :param tags: Tags available for search
        :param added_by: Posts uploaded by specified users
        :param voted: Posts voted by specified user
        :param page_number: Page number from which to start iteration
        :param limit: Maximum amount of posts per page
        """
        async for page in PostPaginator(
            self._http_client, const.POST_URL, **from_locals(locals())
        ):
            for post in page.items:
                yield post

    async def get_favorited_posts(self) -> AsyncIterator[mdl.Post]:
        """Shorthand way to get favorited posts of currently logged-in user."""
        if self._profile is None:
            raise errors.LoginRequirementError

        async for post in self.browse_posts(favorited_by=self._profile.name):
            yield post

    async def get_top_posts(self) -> AsyncIterator[mdl.Post]:
        """Shorthand way to get top posts."""
        async for post in self.browse_posts(order=types.PostOrder.QUALITY):
            yield post

    async def get_popular_posts(self) -> AsyncIterator[mdl.Post]:
        """Shorthand way to get popular posts."""
        async for post in self.browse_posts(order=types.PostOrder.POPULARITY):
            yield post

    async def get_recommended_posts(self) -> AsyncIterator[mdl.Post]:
        """Shorthand way to get recommended posts for the currently logged-in user."""
        if self._profile is None:
            raise errors.LoginRequirementError

        async for post in self.browse_posts(recommended_for=self._profile.name):
            yield post

    async def get_similar_posts(self, post_id: int) -> AsyncIterator[mdl.Post]:
        """Get posts similar (recommended) for specific post by its ID."""
        tag = f"recommended_for_post:{post_id}"
        async for post in self.browse_posts(tags=[tag]):
            yield post

    async def get_post_comments(self, post_id: int) -> AsyncIterator[mdl.Comment]:
        """Get comments of the specific post by its ID."""
        async for page in Paginator(
            self._http_client,
            const.COMMENT_URL.format(post_id=post_id),
            mdl.Comment
        ):
            for comment in page.items:
                yield comment

    async def get_post(self, post_id: int) -> mdl.Post:
        """Get specific post by its ID."""
        response = await self._http_client.get(f"{const.POST_URL}/{post_id}")

        if not response.ok:
            raise errors.PageNotFoundError(response.status, post_id=post_id)

        return mdl.Post(**response.json)

    async def create_post(self):  # TODO: TBA
        raise NotImplementedError


class AIClient(BaseClient):
    """Client for working with Sankaku built-in AI."""
    async def browse_ai_posts(
        self,
        *,
        page_number: Optional[int] = None,
        limit: Optional[Annotated[int, ValueRange(1, 100)]] = None
    ) -> AsyncIterator[mdl.AIPost]:
        """Get AI created posts from AI dedicated post pages.

        :param page_number: Page number from which to start iteration
        :param limit: Maximum amount of posts per page
        """
        async for page in Paginator(
            self._http_client, const.AI_POST_URL,
            mdl.AIPost, **from_locals(locals())
        ):
            for post in page.items:
                yield post

    async def get_ai_post(self, post_id: int) -> mdl.AIPost:
        """Get specific AI post by its ID."""
        response = await self._http_client.get(f"{const.AI_POST_URL}/{post_id}")

        if not response.ok:
            raise errors.PageNotFoundError(response.status, post_id=post_id)

        return mdl.AIPost(**response.json)

    async def create_ai_post(self):  # TODO: TBA
        raise NotImplementedError


class TagClient(BaseClient):
    """Client for tag browsing."""
    async def browse_tags(
        self,
        tag_type: Optional[types.TagType] = None,  # TODO: ability to specify multiple tags
        order: Optional[types.TagOrder] = None,
        rating: Optional[types.Rating] = None,
        max_post_count: Optional[int] = None,
        sort_parameter: Optional[types.SortParameter] = None,
        sort_direction: Optional[types.SortDirection] = None,
        *,
        page_number: Optional[int] = None,
        limit: Optional[Annotated[int, ValueRange(1, 100)]] = None

    ) -> AsyncIterator[mdl.PageTag]:
        """Get tags from tag pages.

        :param tag_type: Tag type filter
        :param order: Tag order rule
        :param rating: Tag rating
        :param max_post_count: Upper threshold for number of posts with tags found
        :param sort_parameter: Tag sorting parameter
        :param sort_direction: Tag sorting direction
        :param page_number: Page number from which to start iteration
        :param limit: Maximum amount of tags per page
        """
        async for page in TagPaginator(
            self._http_client, const.TAG_URL, **from_locals(locals())
        ):
            for tag in page.items:
                yield tag

    async def get_tag(self, name_or_id: str | int) -> mdl.WikiTag:
        """Get specific tag by its name or ID."""
        ref = "/name" if isinstance(name_or_id, str) else "/id"
        url = f"{const.TAG_WIKI_URL}{ref}/{name_or_id}"

        response = await self._http_client.get(url)

        if not response.ok:
            raise errors.PageNotFoundError(response.status, name_or_id=name_or_id)

        return mdl.WikiTag(wiki=response.json["wiki"], **response.json["tag"])


class BookClient(BaseClient):
    """Client for book (pool) browsing."""
    async def browse_books(
        self,
        order: Optional[types.BookOrder] = None,
        rating: Optional[types.Rating] = None,
        recommended_for: Optional[str] = None,
        favorited_by: Optional[str] = None,
        tags: Optional[list[str]] = None,
        added_by: Optional[list[str]] = None,
        voted: Optional[str] = None,
        *,
        page_number: Optional[int] = None,
        limit: Optional[Annotated[int, ValueRange(1, 100)]] = None
    ) -> AsyncIterator[mdl.PageBook]:
        """Get books from book (pool) pages.

        :param order: Book order rule
        :param rating: Books rating
        :param recommended_for: Books recommended for specified user
        :param favorited_by: Books favorited by specified user
        :param tags: Tags available for search
        :param added_by: Books uploaded by specified users
        :param voted: Books voted by specified user
        :param page_number: Page number from which to start iteration
        :param limit: Maximum amount of books per page
        """
        async for page in BookPaginator(
            self._http_client, const.BOOK_URL, **from_locals(locals())
        ):
            for book in page.items:
                yield book

    async def get_favorited_books(self) -> AsyncIterator[mdl.PageBook]:
        """Shorthand way to get favorited books for the currently logged-in user."""
        if self._profile is None:
            raise errors.LoginRequirementError

        async for book in self.browse_books(favorited_by=self._profile.name):
            yield book

    async def get_recommended_books(self) -> AsyncIterator[mdl.PageBook]:
        """Shorthand way to get recommended books for the currently logged-in user."""
        if self._profile is None:
            raise errors.LoginRequirementError

        async for book in self.browse_books(recommended_for=self._profile.name):
            yield book

    async def get_recently_read_books(self) -> AsyncIterator[mdl.PageBook]:
        """Get recently read/opened books of the currently logged-in user."""
        if self._profile is None:
            raise errors.LoginRequirementError

        async for book in self.browse_books(tags=[f"read:@{self._profile.id}@"]):
            yield book

    async def get_related_books(self, post_id: int) -> AsyncIterator[mdl.PageBook]:
        """Get books related to specific post by its ID."""
        async for page in BookPaginator(
            self._http_client, const.RELATED_BOOK_URL.format(post_id=post_id)
        ):
            for book in page.items:
                yield book

    async def get_book(self, book_id: int) -> mdl.Book:
        """Get specific book by its ID."""
        response = await self._http_client.get(f"{const.BOOK_URL}/{book_id}")

        if not response.ok:
            raise errors.PageNotFoundError(response.status, book_id=book_id)

        return mdl.Book(**response.json)


class UserClient(BaseClient):
    """Client for browsing users."""
    async def browse_users(
        self,
        order: Optional[types.UserOrder] = None,
        level: Optional[types.UserLevel] = None,
        *,
        page_number: Optional[int] = None,
        limit: Optional[Annotated[int, ValueRange(1, 100)]] = None
    ) -> AsyncIterator[mdl.User]:
        """Get user profiles from user pages.

        :param order: User order rule
        :param level: User level type
        :param page_number: Page number from which to start iteration
        :param limit: Maximum amount of users per page
        """
        async for page in UserPaginator(
            self._http_client, const.USER_URL, **from_locals(locals())
        ):
            for user in page.items:
                yield user

    async def get_user(self, name_or_id: str | int) -> mdl.User:
        """Get specific user by its name or ID."""
        ref = "/name" if isinstance(name_or_id, str) else ""
        url = f"{const.USER_URL}{ref}/{name_or_id}"

        response = await self._http_client.get(url)

        if not response.ok:
            raise errors.PageNotFoundError(response.status, name_or_id=name_or_id)

        return mdl.User(**response.json)



from abc import ABC, abstractmethod
from collections.abc import AsyncIterator
from typing import Generic, TypeVar

from sankaku import errors, models as mdl


__all__ = ["ABCPaginator"]


_T = TypeVar("_T")


class ABCPaginator(ABC, Generic[_T]):
    """Abstract paginator class."""
    @abstractmethod
    def __init__(self, *args, **kwargs) -> None:
        pass

    def __aiter__(self) -> AsyncIterator[mdl.Page[_T]]:
        return self

    async def __anext__(self) -> mdl.Page[_T]:
        try:
            return await self.next_page()
        except errors.PaginatorLastPage:
            raise StopAsyncIteration

    @abstractmethod
    async def next_page(self) -> mdl.Page[_T]:
        """Get paginator next page."""

    @abstractmethod
    def complete_params(self) -> None:
        """Complete params passed to paginator for further use."""



"""Necessary constants such as hardcoded headers, API urls and endpoints,
default values of parameters etc.
"""

HEADERS: dict[str, str] = {
    "user-agent": (
        "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 "
        "(KHTML, like Gecko) Chrome/94.0.4606.85 YaBrowser/21.11.0.1996 "
        "Yowser/2.5 Safari/537.36"
    ),
    "content-type": "application/json; charset=utf-8",
    "x-requested-with": "com.android.browser",
    "accept-encoding": "gzip, deflate, br",
    "host": "capi-v2.sankakucomplex.com"
}


BASE_URL = "https://login.sankakucomplex.com"
API_URL = "https://capi-v2.sankakucomplex.com"

LOGIN_URL = f"{BASE_URL}/auth/token"
POST_URL = f"{API_URL}/posts"
AI_POST_URL = f"{API_URL}/ai_posts"
TAG_URL = f"{API_URL}/tags"
TAG_WIKI_URL = f"{API_URL}/tag-and-wiki"
BOOK_URL = f"{API_URL}/pools"
USER_URL = f"{API_URL}/users"
PROFILE_URL = f"{USER_URL}/me"

# Not fully completed urls for usage with str.format()
COMMENT_URL = f"{POST_URL}/{{post_id}}/comments"
RELATED_BOOK_URL = f"{API_URL}/post/{{post_id}}/pools"


BASE_RPS = 3
BASE_RPM = 180
BASE_PAGE_NUMBER = 1
BASE_PAGE_LIMIT = 40

PAGE_ALLOWED_ERRORS = [
    "snackbar__anonymous-recommendations-limit-reached",
    "snackbar__account_offset-forbidden"
]

DEFAULT_TOKEN_TYPE = "Bearer"



from typing import Optional, Literal, Annotated, TypeVar
from datetime import datetime

from .abc import ABCPaginator
from sankaku.typedefs import ValueRange
from sankaku.utils import ratelimit
from sankaku import models as mdl, constants as const, types, errors
from sankaku.clients import HttpClient


__all__ = [
    "Paginator",
    "PostPaginator",
    "TagPaginator",
    "BookPaginator",
    "UserPaginator"
]


_T = TypeVar("_T")


class Paginator(ABCPaginator[_T]):
    """Basic paginator for iteration without any special parameters."""
    def __init__(
        self,
        http_client: HttpClient,
        url: str,
        model: type[_T],
        page_number: Optional[int] = None,
        limit: Optional[Annotated[int, ValueRange(1, 100)]] = None,
        params: Optional[dict[str, str]] = None
    ) -> None:
        self.http_client = http_client
        self.url = url
        self.model = model
        self.page_number = page_number or const.BASE_PAGE_NUMBER
        self.limit = limit or const.BASE_PAGE_LIMIT
        self.params: dict[str, str] = params or {}

        self.complete_params()

    @ratelimit(rps=const.BASE_RPS)
    async def next_page(self) -> mdl.Page[_T]:  # type: ignore[override]
        """Get paginator next page."""
        response = await self.http_client.get(self.url, params=self.params)
        match response.json:
            case [] | {"data": []}:
                raise errors.PaginatorLastPage(response.status, page=self.page_number)
            case {"code": code} if code in const.PAGE_ALLOWED_ERRORS:
                raise errors.PaginatorLastPage(response.status, page=self.page_number)
            case {"code": _, "errorId": _}:
                raise errors.SankakuServerError(response.status, **response.json)
            case {"data": list() as data} if data:
                response.json = data

        self.page_number += 1
        self.params["page"] = str(self.page_number)
        return self._construct_page(response.json)

    def complete_params(self) -> None:
        """Complete params passed to paginator for further use."""
        self.params["lang"] = "en"
        if self.page_number is not None:
            self.params["page"] = str(self.page_number)
        if self.limit is not None:
            self.params["limit"] = str(self.limit)

    def _construct_page(self, data: list[dict]) -> mdl.Page[_T]:
        """Construct and return page model."""
        items = [self.model(**d) for d in data]
        return mdl.Page[_T](number=self.page_number - 1, items=items)


class PostPaginator(Paginator[mdl.Post]):
    """Paginator used for iteration through post pages."""
    def __init__(
        self,
        http_client: HttpClient,
        url: str,
        model: type[mdl.Post] = mdl.Post,
        page_number: Optional[int] = None,
        limit: Optional[Annotated[int, ValueRange(1, 100)]] = None,
        params: Optional[dict[str, str]] = None,
        order: Optional[types.PostOrder] = None,
        date: Optional[list[datetime]] = None,
        rating: Optional[types.Rating] = None,
        threshold: Optional[Annotated[int, ValueRange(1, 100)]] = None,
        hide_posts_in_books: Optional[Literal["in-larger-tags", "always"]] = None,
        file_size: Optional[types.FileSize] = None,
        file_type: Optional[types.FileType] = None,
        video_duration: Optional[list[int]] = None,
        recommended_for: Optional[str] = None,
        favorited_by: Optional[str] = None,
        tags: Optional[list[str]] = None,
        added_by: Optional[list[str]] = None,
        voted: Optional[str] = None
    ) -> None:
        self.order = order
        self.date = date
        self.rating = rating
        self.threshold = threshold
        self.hide_posts_in_books = hide_posts_in_books
        self.file_size = file_size
        self.file_type = file_type
        self.video_duration = video_duration
        self.recommended_for = recommended_for
        self.favorited_by = favorited_by
        self.tags = tags
        self.added_by = added_by
        self.voted = voted
        super().__init__(http_client, url, model, page_number, limit, params)

    def complete_params(self) -> None:
        super().complete_params()
        if self.tags is None:
            self.tags = []

        for items in self.__dict__.items():
            match items:
                case [_, None]:
                    continue
                case ["order" | "rating" | "file_type" as k, v] if v is not types.FileType.IMAGE:  # noqa
                    self.tags.append(f"{k}:{v.value}")
                case ["threshold" | "recommended_for" | "voted" as k, v]:
                    self.tags.append(f"{k}:{v}")
                case ["file_size", _]:
                    self.tags.append(self.file_size.value)  # type: ignore[union-attr]
                case ["date", _]:
                    date = "..".join(d.strftime("%Y-%m-%dT%H:%M") for d in self.date)  # type: ignore[union-attr]
                    self.tags.append(f"date:{date}")
                case ["video_duration", _] if self.file_type is not types.FileType.VIDEO:  # noqa
                    raise errors.VideoDurationError
                case ["video_duration", _]:
                    duration = "..".join(str(sec) for sec in self.video_duration)  # type: ignore[union-attr]
                    self.tags.append(f"duration:{duration}")
                case ["favorited_by", _]:
                    self.tags.append(f"fav:{self.favorited_by}")
                case ["added_by", _]:
                    for user in self.added_by:  # type: ignore[union-attr]
                        self.tags.append(f"user:{user}")

        if self.hide_posts_in_books is not None:
            self.params["hide_posts_in_books"] = self.hide_posts_in_books
        if self.tags:
            self.params["tags"] = " ".join(self.tags)


class TagPaginator(Paginator[mdl.PageTag]):
    """Paginator used for iteration through tag pages."""
    def __init__(
        self,
        http_client: HttpClient,
        url: str,
        model: type[mdl.PageTag] = mdl.PageTag,
        page_number: Optional[int] = None,
        limit: Optional[Annotated[int, ValueRange(1, 100)]] = None,
        params: Optional[dict[str, str]] = None,
        tag_type: Optional[types.TagType] = None,
        order: Optional[types.TagOrder] = None,
        rating: Optional[types.Rating] = None,
        max_post_count: Optional[int] = None,
        sort_parameter: Optional[types.SortParameter] = None,
        sort_direction: Optional[types.SortDirection] = None
    ) -> None:
        self.tag_type = tag_type
        self.order = order
        self.rating = rating
        self.max_post_count = max_post_count
        self.sort_parameter = sort_parameter
        self.sort_direction = sort_direction or types.SortDirection.DESC
        super().__init__(http_client, url, model, page_number, limit, params)

    def complete_params(self) -> None:
        super().complete_params()
        if self.tag_type is not None:
            self.params["types[]"] = str(self.tag_type.value)
        if self.order is not None:
            self.params["order"] = self.order.value
        if self.rating is not None:
            self.params["rating"] = self.rating.value
        if self.max_post_count is not None:
            self.params["amount"] = str(self.max_post_count)
        if self.sort_parameter is not None:
            self.params.update(
                sortBy=self.sort_parameter.value,
                sortDirection=self.sort_direction.value
            )


class BookPaginator(Paginator[mdl.PageBook]):
    """Paginator used for iteration through book pages."""
    def __init__(
        self,
        http_client: HttpClient,
        url: str,
        model: type[mdl.PageBook] = mdl.PageBook,
        page_number: Optional[int] = None,
        limit: Optional[Annotated[int, ValueRange(1, 100)]] = None,
        params: Optional[dict[str, str]] = None,
        order: Optional[types.BookOrder] = None,
        rating: Optional[types.Rating] = None,
        recommended_for: Optional[str] = None,
        favorited_by: Optional[str] = None,
        tags: Optional[list[str]] = None,
        added_by: Optional[list[str]] = None,
        voted: Optional[str] = None
    ) -> None:
        self.order = order
        self.rating = rating
        self.recommended_for = recommended_for
        self.favorited_by = favorited_by
        self.tags = tags
        self.added_by = added_by
        self.voted = voted
        super().__init__(http_client, url, model, page_number, limit, params)

    def complete_params(self) -> None:
        super().complete_params()
        if self.tags is None:
            self.tags = []

        for items in self.__dict__.items():
            match items:
                case [_, None]:
                    continue
                case ["order" | "rating" as k, v]:  # noqa
                    self.tags.append(f"{k}:{v.value}")
                case ["recommended_for" | "voted" as k, v]:
                    self.tags.append(f"{k}:{v}")
                case ["favorited_by", _]:
                    self.tags.append(f"fav:{self.favorited_by}")
                case ["added_by", _]:
                    for user in self.added_by:  # type: ignore[union-attr]
                        self.tags.append(f"user:{user}")

        if self.tags:
            self.params["tags"] = " ".join(self.tags)


class UserPaginator(Paginator[mdl.User]):
    """Paginator used for iteration through user pages."""
    def __init__(
        self,
        http_client: HttpClient,
        url: str,
        model: type[mdl.User] = mdl.User,
        page_number: Optional[int] = None,
        limit: Optional[Annotated[int, ValueRange(1, 100)]] = None,
        params: Optional[dict[str, str]] = None,
        order: Optional[types.UserOrder] = None,
        level: Optional[types.UserLevel] = None
    ) -> None:
        self.order = order
        self.level = level
        super().__init__(http_client, url, model, page_number, limit, params)

    def complete_params(self) -> None:
        super().complete_params()
        if self.order is not None:
            self.params["order"] = self.order.value
        if self.level is not None:
            self.params["level"] = str(self.level.value)



from abc import ABC, abstractmethod
from typing import Optional, Self

from sankaku.models.http import ClientResponse


__all__ = ["ABCHttpClient", "ABCClient"]


class ABCHttpClient(ABC):
    """Abstract client for handling http requests."""
    @abstractmethod
    def __init__(self, *args, **kwargs) -> None:
        pass

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.close()

    @abstractmethod
    def __del__(self) -> None:
        pass

    @abstractmethod
    async def close(self) -> None:
        """Close previously created client session."""

    @abstractmethod
    async def request(self, method: str, url: str, **kwargs) -> ClientResponse:
        """Make request to specified url."""


class ABCClient(ABC):
    """Abstract Sankaku client."""
    @abstractmethod
    def __init__(self, *args, **kwargs) -> None:
        pass

    @abstractmethod
    async def login(
        self,
        *,
        access_token: Optional[str] = None,
        login: Optional[str] = None,
        password: Optional[str] = None
    ) -> None:
        """Login into sankakucomplex.com via access token or credentials."""



from enum import Enum


__all__ = [
    "Rating",
    "PostOrder",
    "SortParameter",
    "SortDirection",
    "TagOrder",
    "TagType",
    "FileType",
    "FileSize",
    "UserOrder",
    "UserLevel",
    "BookOrder"
]


class Rating(str, Enum):
    SAFE = "s"
    QUESTIONABLE = "q"
    EXPLICIT = "e"


class PostOrder(Enum):
    POPULARITY = "popularity"
    DATE = "date"
    QUALITY = "quality"
    RANDOM = "random"
    RECENTLY_FAVORITED = "recently_favorited"
    RECENTLY_VOTED = "recently_voted"


class SortParameter(Enum):
    NAME = "name"
    TRANSLATIONS = "name_ja"
    TYPE = "type"
    RATING = "rating"
    BOOK_COUNT = "pool_count"
    POST_COUNT = "count"


class SortDirection(Enum):
    ASC = "asc"
    DESC = "desc"


class TagOrder(Enum):
    POPULARITY = "popularity"
    QUALITY = "quality"


class TagType(Enum):
    ARTIST = 1
    COPYRIGHT = 3
    CHARACTER = 4
    GENERAL = 0

    MEDIUM = 8
    META = 9
    GENRE = 5
    STUDIO = 2


class FileType(Enum):
    IMAGE = "image"  # jpeg, png, webp formats
    GIF = "animated_gif"  # gif format
    VIDEO = "video"  # mp4, webm formats


class FileSize(Enum):
    LARGE = "large_filesize"
    HUGE = "extremely_large_filesize"
    LONG = "long_image"
    WALLPAPER = "wallpaper"
    A_RATIO_16_9 = "16:9_aspect_ratio"
    A_RATIO_4_3 = "4:3_aspect_ratio"
    A_RATIO_3_2 = "3:2_aspect_ratio"
    A_RATIO_1_1 = "1:1_aspect_ratio"


class UserOrder(Enum):
    POSTS = "post_upload_count"
    FAVORITES = "favorite_count"
    NAME = "name"
    NEWEST = "newest"
    OLDEST = "oldest"
    LAST_SEEN = "active"


class UserLevel(Enum):
    ADMIN = 50
    SYSTEM_USER = 45
    MODERATOR = 40
    JANITOR = 35
    CONTRIBUTOR = 33
    PRIVILEGED = 30
    MEMBER = 20
    BLOCKED = 10
    UNACTIVATED = 0


class BookOrder(Enum):
    POPULARITY = "popularity"
    DATE = "date"
    QUALITY = "quality"
    RANDOM = "random"
    RECENTLY_FAVORITED = "recently_favorited"
    RECENTLY_VOTED = "recently_voted"



from .http_client import HttpClient
from .clients import *


__all__ = ["HttpClient", "SankakuClient"]


class SankakuClient(  # noqa
    PostClient,
    AIClient,
    TagClient,
    BookClient,
    UserClient
):
    """Simple client for Sankaku API."""



from __future__ import annotations

from typing import Optional
from datetime import datetime

from pydantic import BaseModel

from sankaku import types
from .users import Author
from .tags import PostTag
from .posts import Post


__all__ = ["PageBook", "Book"]


class BookState(BaseModel):
    current_page: int
    sequence: int
    post_id: int
    series_id: Optional[int]
    created_at: datetime
    updated_at: datetime
    percent: int


class PageBook(BaseModel):
    """Model that describes books on book pages."""
    id: int
    name_en: Optional[str]
    name_ja: Optional[str]
    description: str
    description_en: Optional[str]
    description_ja: Optional[str]
    created_at: datetime
    updated_at: datetime  # Not sure that update dt isn't optional
    author: Author
    is_public: bool
    is_active: bool
    is_flagged: bool
    post_count: int
    pages_count: int
    visible_post_count: int
    is_intact: bool
    rating: Optional[types.Rating]
    parent_id: Optional[int]
    has_children: Optional[bool]
    is_rating_locked: bool
    fav_count: int
    vote_count: int
    total_score: int
    comment_count: Optional[int]
    tags: list[PostTag]
    post_tags: list[PostTag]
    artist_tags: list[PostTag]
    genre_tags: list[PostTag]
    is_favorited: bool
    user_vote: Optional[int]
    posts: list[Post]
    file_url: Optional[str]
    sample_url: Optional[str]
    preview_url: Optional[str]
    cover_post: Optional[Post]
    reading: Optional[BookState]
    is_premium: bool
    is_pending: bool
    is_raw: bool
    is_trial: bool
    redirect_to_signup: bool
    locale: str
    is_deleted: bool
    cover_post_id: Optional[int]
    name: Optional[str]
    parent_pool: Optional[PageBook]


class Book(PageBook):
    """Model that describes specific book."""
    child_pools: Optional[list[PageBook]]
    flagged_by_user: bool
    prem_post_count: int



import setuptools

setuptools.setup(
    name="sankaku",
    version="1.0.0",
    author="zerex290",
    author_email="zerex290@gmail.com",
    description="Asynchronous API wrapper for Sankaku Complex.",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    keywords="sankaku sankakucomplex api".split(),
    url="https://github.com/zerex290/sankaku",
    project_urls={"Issue Tracker": "https://github.com/zerex290/sankaku/issues"},
    packages=setuptools.find_packages(exclude=["tests", "tests.*"]),
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.11",
    install_requires=["aiohttp", "pydantic", "loguru"]
)



from dataclasses import dataclass
from typing import Any


__all__ = ["ClientResponse"]


@dataclass()
class ClientResponse:
    """Dataclass that preserves information from aiohttp ClientResponse."""
    status: int
    json: Any

    @property
    def ok(self) -> bool:
        """Check if response status code is less than 400."""
        return self.status < 400
